import colorama


def resource_checking(title: str):
    def _resource_checking(func):
        def wrapper(*args, **kwargs):
            from data_to_paper.utils.highlighted_text import colored_text
            from data_to_paper.utils.print_to_file import IS_LOGGING_ENABLED
            print(f"\n*** {title}\nRunning...")
            try:
                with IS_LOGGING_ENABLED.temporary_set(False):
                    func(*args, **kwargs)
            except Exception as e:
                print(colored_text(f"Test failed:\n{e}", color=colorama.Fore.RED))
                return
            print(colored_text("Test successful!", color=colorama.Fore.GREEN))
        return wrapper
    return _resource_checking



from __future__ import annotations

import os
import shutil
import zipfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, List, Union

import pandas as pd

from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.env import FOLDER_FOR_RUN
from data_to_paper.latex.clean_latex import wrap_as_latex_code_output
from data_to_paper.utils.file_utils import run_in_directory, clear_directory
from data_to_paper.utils.mutable import Mutable
from data_to_paper.code_and_output_files.referencable_text import NumericReferenceableText, \
    hypertarget_if_referencable_text_product, ReferencableTextProduct

TEXT_EXTS = ['.txt', '.md', '.csv', '.xls', '.xlsx']


@dataclass(frozen=True)
class DataFileDescription:
    file_path: str  # relative to the data directory.  should normally just be the file name
    description: Optional[Union[str, ReferencableTextProduct]] = None  # a user provided description of the file
    originated_from: Optional[str] = None  # None for raw file
    is_binary: Optional[bool] = None  # None for auto based on file extension

    def get_is_binary(self):
        """
        Return True if the file is binary.
        """
        if self.is_binary is not None:
            return self.is_binary

        return Path(self.file_path).suffix not in TEXT_EXTS

    def is_excel(self):
        return Path(self.file_path).suffix in ['.xlsx', '.xls']

    def get_file_header(self, num_lines: int = 4):
        """
        Return the first `num_lines` lines of the file (if they exist).
        """
        if self.is_excel():
            # go over all sheets and return all of them:
            df = pd.read_excel(self.file_path, sheet_name=None)
            s = f'This is an Excel file with {len(df)} sheets. Here is the first few rows for each sheet:\n\n'
            for sheet_name in df.keys():
                s += f'### Sheet: "{sheet_name}"\n'
                s += f'```output\n{df[sheet_name].head(num_lines).to_string(index=False)}\n```\n'
            s += '\n'
            return s

        with open(self.file_path) as f:
            head = []
            for _ in range(num_lines):
                try:
                    head.append(next(f))
                except StopIteration:
                    break
                except UnicodeDecodeError:
                    head.append('UnicodeDecodeError\n')
            return f'Here are the first few lines of the file:\n' \
                   f'```output\n{"".join(head)}\n```\n'

    def pretty_repr(self, num_lines: int = 4, view_purpose: ViewPurpose = None, file_num: Optional[int] = None) -> str:
        if file_num is not None:
            s = f'### File {file_num}: "{self.file_path}"\n'
        else:
            s = f'### "{self.file_path}"\n'
        description = self.description
        if description is not None:
            description = hypertarget_if_referencable_text_product(description, view_purpose, with_header=False)
            s += f'{description}\n\n'
        if num_lines > 0 and not self.get_is_binary():
            s += self.get_file_header(num_lines)
        return s


class DataFileDescriptions(List[DataFileDescription]):
    """
    A list of data file descriptions.
    """

    def __init__(self, *args, data_folder: Optional[Union[str, Path]] = None,
                 general_description: Optional[Union[str, ReferencableTextProduct]] = None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.data_folder = data_folder
        self.general_description = general_description

    @classmethod
    def from_other(cls, other: DataFileDescriptions):
        return cls(other, data_folder=other.data_folder, general_description=other.general_description)

    def __str__(self):
        return self.pretty_repr()

    def get_file_description(self, file_path: str) -> DataFileDescription:
        """
        Return the data file description for the given file path.
        """
        for data_file_description in self:
            if data_file_description.file_path == file_path:
                return data_file_description
        raise ValueError(f"Could not find file path {file_path} in data file descriptions.")

    def get_children(self, data_file: DataFileDescription):
        """
        Return the children of the given data file.
        """
        return DataFileDescriptions([data_file_description for data_file_description in self
                                     if data_file_description.originated_from == data_file.file_path],
                                    data_folder=self.data_folder)

    def get_all_raw_files(self):
        """
        Return all the raw files.
        """
        return DataFileDescriptions([data_file_description for data_file_description in self
                                     if data_file_description.originated_from is None],
                                    data_folder=self.data_folder)

    def get_pretty_description_for_file_and_children(
            self, data_file: DataFileDescription, index: Mutable = None, view_purpose: ViewPurpose = None):
        """
        Return a pretty description for the given data file and all its children.
        """
        children = self.get_children(data_file)
        index.val += 1
        s = data_file.pretty_repr(0 if children else 4, view_purpose=view_purpose, file_num=index.val) + '\n'
        for child in children:
            s += self.get_pretty_description_for_file_and_children(child, index)
        return s

    def pretty_repr(self, num_lines: int = 4, view_purpose: ViewPurpose = None) -> str:
        s = ''
        if self.general_description is not None:
            s += '## General Description\n'
            s += hypertarget_if_referencable_text_product(self.general_description, view_purpose,
                                                          with_header=False) + '\n'
        with run_in_directory(self.data_folder):
            s += '## Data Files\n'
            if len(self) == 0:
                s += 'There are no data files'
            elif len(self) == 1:
                s += f"The dataset consists of 1 data file:\n\n"
                s += self[0].pretty_repr(num_lines, view_purpose=view_purpose)
            else:
                s += f"The dataset consists of {len(self)} data files:\n\n"
                index = Mutable(0)
                for parent in self.get_all_raw_files():
                    s += self.get_pretty_description_for_file_and_children(parent, index, view_purpose=view_purpose)
            return s

    def get_data_filenames(self):
        return [data_file_description.file_path for data_file_description in self]

    def to_latex(self,
                 section_name: str = 'Data Description',
                 text: str = 'Here is the data description, as provided by the user:',
                 view_purpose: ViewPurpose = None) -> str:
        s = ''
        s += f"\\section{{{section_name}}}\n{text}\n\n"
        s += wrap_as_latex_code_output(
            self.pretty_repr(num_lines=0, view_purpose=view_purpose))
        return s


@dataclass
class CreateDataFileDescriptions:
    """
    Given a project directory, create a temp folder to run in, and
    the data file descriptions.
    """
    project_directory: Path = None
    data_files_str_paths: List[str] = field(default_factory=list)
    data_files_is_binary: List[Optional[bool]] = field(default_factory=list)

    GENERAL_DESCRIPTION_FILENAME = 'general_description.txt'
    DESCRIPTION_FILENAME_EXT = '.description.txt'

    GENERAL_FILE_DESCRIPTION_PREFIX = 'S'
    FILE_DESCRIPTIONS_PREFIXES = ('T', 'U', 'V', 'W', 'X', 'Y', 'Z')

    temp_folder_to_run_in: Path = FOLDER_FOR_RUN

    def _get_description_file_path(self, data_file_path_str: str):
        data_file_path = self._convert_data_file_path_str_to_path(data_file_path_str)
        return self.project_directory / (data_file_path.name + self.DESCRIPTION_FILENAME_EXT)

    def _read_file_description(self, data_file_path_str: str):
        description_file_path = self._get_description_file_path(data_file_path_str)
        if not description_file_path.exists():
            raise FileNotFoundError(f"Description file for {data_file_path_str} not found.")
        return description_file_path.read_text()

    def _read_general_description(self):
        general_description_file_path = self.project_directory / self.GENERAL_DESCRIPTION_FILENAME
        if not general_description_file_path.exists():
            raise FileNotFoundError(f"General description file not found:\n{general_description_file_path}.")
        return general_description_file_path.read_text()

    def _get_hypertarget_prefix(self, file_num: Optional[int] = None, file_name: str = None):
        """
        Return the hypertarget prefix for the given file number.
        file_num: the file number. None refers to the general description.
        """
        return self.GENERAL_FILE_DESCRIPTION_PREFIX if file_num is None \
            else self.FILE_DESCRIPTIONS_PREFIXES[file_num]

    def _convert_description_to_referenceable_text(self, description: str, file_num: Optional[int],
                                                   file_name: str = None):
        """
        file_num: the file number. None refers to the general description.
        """
        return ReferencableTextProduct(
            referencable_text=NumericReferenceableText(
                text=description,
                hypertarget_prefix=self._get_hypertarget_prefix(file_num, file_name)),
            name=file_name,
        )

    def _convert_data_file_path_str_to_path(self, data_file_path: str):
        """
        convert to absolute path.
        if relative path, it is relative to the project directory.
        """
        return Path(data_file_path).absolute() if Path(data_file_path).is_absolute() \
            else self.project_directory / data_file_path

    def _get_file_description_referenceable_text(self, data_file_str_path: str, file_num: Optional[int]):
        description = self._read_file_description(data_file_str_path)
        data_file_path = self._convert_data_file_path_str_to_path(data_file_str_path)
        return self._convert_description_to_referenceable_text(description, file_num, data_file_path.name)

    def _get_general_description_referenceable_text(self):
        return self._convert_description_to_referenceable_text(self._read_general_description(), None,
                                                               'General Description')

    def _copy_files_and_get_list_of_data_file_descriptions(self) -> List[DataFileDescription]:
        data_file_descriptions = []
        clear_directory(self.temp_folder_to_run_in)  # clear data folder
        for j, data_file_str_path in enumerate(self.data_files_str_paths):
            data_file_path = self._convert_data_file_path_str_to_path(data_file_str_path)
            data_file_path_zip = data_file_path.with_name(data_file_path.name + '.zip')
            if os.path.exists(data_file_path):
                # copy file to data folder
                shutil.copyfile(data_file_path, self.temp_folder_to_run_in / data_file_path.name)
            elif os.path.exists(data_file_path_zip):
                # unzip file to data folder
                with zipfile.ZipFile(data_file_path_zip, 'r') as zip_ref:
                    zip_ref.extractall(self.temp_folder_to_run_in)
            else:
                raise FileNotFoundError(f"File {data_file_path.name} or {data_file_path.name}.zip "
                                        f"not found in {data_file_path.parent}")

            data_file_descriptions.append(DataFileDescription(
                file_path=data_file_path.name,  # relative to data dir
                description=self._get_file_description_referenceable_text(data_file_str_path, j),
                is_binary=self.data_files_is_binary[j]
            ))
        return data_file_descriptions

    def create_temp_folder_and_get_file_descriptions(self) -> DataFileDescriptions:
        file_descriptions = self._copy_files_and_get_list_of_data_file_descriptions()
        return DataFileDescriptions(
            file_descriptions,
            data_folder=self.temp_folder_to_run_in,
            general_description=self._get_general_description_referenceable_text()
        )

    def get_raw_str_data_file_descriptions(self) -> DataFileDescriptions:
        return DataFileDescriptions(
            [DataFileDescription(
                file_path=data_file_str_path,  # relative to data dir
                description=self._read_file_description(data_file_str_path),
                is_binary=self.data_files_is_binary[j]
            ) for j, data_file_str_path in enumerate(self.data_files_str_paths)],
            data_folder=None,
            general_description=self._read_general_description(),
        )

    def create_file_descriptions(self, general_description: str, data_file_descriptions: List[str]):
        """
        Create the file descriptions.
        """
        (self.project_directory / self.GENERAL_DESCRIPTION_FILENAME).write_text(general_description)
        for j, data_file_str_path in enumerate(self.data_files_str_paths):
            self._get_description_file_path(data_file_str_path).write_text(data_file_descriptions[j])

    def check_files_exist(self):
        for data_file_str_path in self.data_files_str_paths:
            data_file_path = self._convert_data_file_path_str_to_path(data_file_str_path)
            data_file_path_zip = data_file_path.with_name(data_file_path.name + '.zip')
            if not data_file_path.exists() and not data_file_path_zip.exists():
                raise FileNotFoundError(f"Data file {data_file_path} not found.")



"""
Run this script to check if the API keys for the LLMServer and Semantic Scholar are valid.
"""
from data_to_paper.latex.check_dependencies import check_all_pdflatex_dependencies
from data_to_paper.latex.latex_to_html import check_pandoc_is_installed
from data_to_paper.servers.check_connection import check_all_servers


def check_resources():
    print('\n\nChecking API keys and correct installation of all dependencies...')
    check_all_servers()
    check_all_pdflatex_dependencies()
    check_pandoc_is_installed()



import os
import re
import subprocess

from data_to_paper.exceptions import MissingInstallationError
from data_to_paper.latex.clean_latex import process_latex_text_and_math
from data_to_paper.utils.file_utils import run_in_temp_directory
from data_to_paper.utils.resource_checking import resource_checking
from data_to_paper.utils.text_formatting import escape_html


@resource_checking("Checking Pandoc installation")
def check_pandoc_is_installed():
    # This will raise MissingInstallationError:
    raise_if_pandoc_is_not_installed()


def raise_if_pandoc_is_not_installed():
    try:
        subprocess.run(['pandoc', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
    except FileNotFoundError:
        raise MissingInstallationError(package_name="Pandoc", instructions="See: https://pandoc.org/installing.html")


def convert_latex_to_html(latex: str) -> str:
    """
    Convert LaTeX text to HTML using Pandoc through pypandoc.

    Parameters:
    - latex (str): A string containing LaTeX code.

    Returns:
    - str: The converted HTML text.
    """
    raise_if_pandoc_is_not_installed()

    is_title = re.search(pattern=r'\\title{(.+?)}', string=latex)
    dir_path = os.path.dirname(os.path.realpath(__file__))
    tex_file = 'temp.tex'
    command = [
        'pandoc', '-s', tex_file,
        '--lua-filter', os.path.join(dir_path, 'adjust_section.lua'),
        '-t', 'html'
    ]

    # Get the html template:
    if is_title:
        template_name = 'html_template_for_title_latex.html'
    else:
        template_name = 'html_template_for_titleless_latex.html'
    template_path = os.path.join(dir_path, template_name)
    command += ['--template', template_path]

    if not is_title:
        command += ['--metadata', 'title=Titleless LaTeX Document']

    # To show citation commands (like '\\cite{ref1}'):
    command += ['--citeproc']

    try:
        with run_in_temp_directory():
            # process latex and escape special characters
            latex = process_latex_text_and_math(latex)
            # Write the LaTeX into a temporary file
            with open(tex_file, 'w') as f:
                f.write(latex)
            # Convert using Pandoc
            html_output = subprocess.check_output(command, universal_newlines=True)
            return html_output
    except subprocess.CalledProcessError:
        # In case of an error, return the raw latex with proper escaping for HTML
        return escape_html(latex)



from copy import copy

import numpy as np

from dataclasses import dataclass, field
from typing import Optional, Dict, List, Iterable, NamedTuple

from data_to_paper.base_products.product import ValueProduct
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.utils.iterators import interleave
from data_to_paper.utils.mutable import Flag
from data_to_paper.utils.nice_list import NiceList
from data_to_paper.servers.custom_types import Citation


def unite_citation_lists(citations_lists: Iterable[Iterable[Citation]], total: int = None) -> List[Citation]:
    """
    Unite the two lists maintaining a single bibtex_id for each citation.
    Citation that appears in both list will have the query as a set of both quereis.
    """
    bibtex_ids_to_citations = {}
    for citation in interleave(*citations_lists):
        bibtex_id = citation.bibtex_id
        query = citation.query
        if isinstance(query, str):
            query = {query}
        if bibtex_id not in bibtex_ids_to_citations:
            citation = copy(citation)
            citation.query = query
            bibtex_ids_to_citations[bibtex_id] = citation
        else:
            bibtex_ids_to_citations[bibtex_id].query.update(query)
            bibtex_ids_to_citations[bibtex_id].search_rank = min(bibtex_ids_to_citations[bibtex_id].search_rank,
                                                                 citation.search_rank)
        if total is not None:
            if len(bibtex_ids_to_citations) == total:
                break
    return list(bibtex_ids_to_citations.values())


CITATION_REPR_FIELDS_FOR_LLM = \
    ('bibtex_id', 'title', 'journal_and_year', 'tldr', 'influence')
CITATION_REPR_FIELDS_FOR_PRINT = \
    ('query', 'search_rank', 'bibtex_id', 'title', 'journal_and_year', 'tldr', 'influence', 'embedding_similarity')
GET_LITERATURE_SEARCH_FOR_PRINT = Flag(False)


@dataclass
class LiteratureSearchQueriesProduct(ValueProduct):
    name: str = "Literature Search Queries"
    value: Dict[str, List[str]] = None

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        s = ''
        level_str = '#' * (level + 1)
        for scope, queries in self.value.items():
            s += f'{level_str} {scope.title()}\n'
            for query in queries:
                s += f'- "{query}"\n'
        return s


@dataclass
class CitationCollectionProduct(ValueProduct):
    name: str = "Citation List"
    value: List[Citation] = None

    def __iter__(self) -> Iterable[Citation]:
        return super().__iter__()

    def _get_citations_as_str(self, is_html: bool, style: str = None,
                              embedding_target: Optional[np.ndarray] = None) -> str:
        return '\n'.join(citation.pretty_repr(
            fields=CITATION_REPR_FIELDS_FOR_LLM if style == 'llm' else CITATION_REPR_FIELDS_FOR_PRINT,
            is_html=is_html,
            embedding_target=embedding_target,
        ) for citation in self.value)

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, style: str = None,
                                       embedding_target: Optional[np.ndarray] = None, **kwargs: str) -> str:
        return self._get_citations_as_str(is_html=False, style=style, embedding_target=embedding_target)

    def _get_content_as_html(self, level: int, style: str = None,
                             embedding_target: Optional[np.ndarray] = None):
        return self._get_citations_as_str(is_html=True, style=style)


@dataclass
class QueryCitationCollectionProduct(CitationCollectionProduct):
    name: str = "Query Citation List"
    query: str = None

    def get_header(self, view_purpose: ViewPurpose = ViewPurpose.PRODUCT, **kwargs) -> str:
        return f'{self.name} for "{self.query}"'


@dataclass
class SortedCitationCollectionProduct(CitationCollectionProduct):
    name: str = "Sorted Citation List"
    scope: Optional[str] = None
    query: Optional[str] = None
    total: int = None
    distribution_factor: Optional[float] = None
    sort_by_similarity: bool = False
    minimal_influence: int = 0

    def get_header(self, view_purpose: ViewPurpose = ViewPurpose.PRODUCT, style: str = 'llm', **kwargs) -> str:
        s = super().get_header()
        s += f', Scope: "{self.scope}"' if self.scope is not None else ''
        s += f', Query: "{self.query}"' if self.query is not None else ''
        if style != 'llm':
            s += f', Total: {self.total}' if self.total is not None else ''
            s += f', Distribution Factor: {self.distribution_factor}' if self.distribution_factor is not None else ''
            s += f', Sort by Similarity: {self.sort_by_similarity}' if self.sort_by_similarity else ''
            s += f', Minimal Influence: {self.minimal_influence}' if self.minimal_influence > 0 else ''
        return s


class LiteratureSearchParams(NamedTuple):
    total: int
    minimal_influence: int
    distribution_factor: Optional[float]
    sort_by_similarity: bool

    def to_dict(self) -> dict:
        return self._asdict()


@dataclass
class LiteratureSearch(ValueProduct):
    # value is scopes_to_queries_to_citations
    value: Dict[str, Dict[str, CitationCollectionProduct]] = field(default_factory=dict)
    embedding_target: Optional[np.ndarray] = None
    scopes_to_search_params: Dict[str, LiteratureSearchParams] = field(default_factory=dict)

    def get_queries(self, scope: Optional[str] = None) -> List[str]:
        """
        Return the queries in the given scope.
        if scope=None, return all queries.
        """
        if scope is None:
            queries = sum([self.get_queries(scope) for scope in self], [])
        else:
            queries = list(self[scope].keys())
        return NiceList(queries, wrap_with='"', separator='\n', prefix='\n', suffix='\n')

    def get_citations(self, scope: Optional[str] = None, query: Optional[str] = None,
                      total: int = None, distribution_factor: Optional[float] = None,
                      sort_by_similarity: bool = False, minimal_influence: int = 0) -> CitationCollectionProduct:
        """
        Return the citations in the given scope.
        If embedding_target is not None, sort the citations by embedding similarity.
        If total is not None, return only the first total citations.
        If total < 0, return only the last total citations.
        """
        if scope is None:
            assert query is None
            citations = unite_citation_lists((iter(self.get_citations(
                    scope=scope,
                    total=int(total / len(self) * distribution_factor) + 1
                    if distribution_factor and total is not None else None,
                    minimal_influence=minimal_influence,
                    distribution_factor=distribution_factor,
                    sort_by_similarity=sort_by_similarity,
            )) for scope in self))
        elif query is None:
            citations = unite_citation_lists((iter(self.get_citations(
                    scope=scope,
                    query=query,
                    total=int(total / len(self[scope]) * distribution_factor) + 1
                    if distribution_factor and total is not None else None,
                    minimal_influence=minimal_influence,
                    distribution_factor=distribution_factor,
                    sort_by_similarity=sort_by_similarity,
                )) for query in self[scope]))
        else:
            citations = self[scope][query]
        citations = list(citations)

        if minimal_influence > 0:
            citations = [citation for citation in citations if citation.influence >= minimal_influence]

        if sort_by_similarity and self.embedding_target is not None:
            citations = sorted(citations, key=lambda citation: citation.get_embedding_similarity(self.embedding_target),
                               reverse=True)
        else:
            citations = sorted(citations, key=lambda citation: citation.search_rank)

        if total is None:
            pass
        elif total < 0:
            citations = citations[total:]
        else:
            citations = citations[:total]
        return SortedCitationCollectionProduct(value=citations, scope=scope, query=query,
                                               total=total, distribution_factor=distribution_factor,
                                               sort_by_similarity=sort_by_similarity,
                                               minimal_influence=minimal_influence)

    def pretty_repr(self, with_scope_and_queries: bool = False,
                    total: int = None, distribution_factor: Optional[float] = None,
                    sort_by_similarity: bool = False,
                    minimal_influence: int = 0,
                    style: str = None,
                    ) -> str:
        s = ''
        for scope in self:
            if with_scope_and_queries:
                s += '\n\n'
                s += f'Scope: {repr(scope)}\n'
                s += f'Queries: {repr(self.get_queries(scope))}\n'
            s += self.pretty_repr_for_scope_and_query(scope=scope,
                                                      total=total // len(self) + 1,
                                                      distribution_factor=distribution_factor,
                                                      sort_by_similarity=sort_by_similarity,
                                                      minimal_influence=minimal_influence,
                                                      style=style)
        return s

    def pretty_repr_for_scope_and_query(self, scope: str, query: Optional[str] = None,
                                        total: int = None, distribution_factor: Optional[float] = None,
                                        sort_by_similarity: bool = False,
                                        minimal_influence: int = 0,
                                        style: str = None) -> str:
        """
        style: 'llm', 'print', 'html'
        """
        style = style or 'llm'
        citations = self.get_citations(scope=scope, query=query, total=total,
                                       distribution_factor=distribution_factor,
                                       sort_by_similarity=sort_by_similarity,
                                       minimal_influence=minimal_influence,
                                       )
        if GET_LITERATURE_SEARCH_FOR_PRINT:
            style = 'print'
        return '\n'.join(citation.pretty_repr(
            fields=CITATION_REPR_FIELDS_FOR_LLM if style == 'llm' else CITATION_REPR_FIELDS_FOR_PRINT,
            is_html=style == 'html',
            embedding_target=self.embedding_target,
        ) for citation in citations)

    def get_header(self, view_purpose: ViewPurpose = ViewPurpose.PRODUCT, scope: Optional[str] = None, **kwargs) -> str:
        return f'{scope}-related literature search' if scope is not None else self.name

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose,
                                       scope: Optional[str] = None, style: str = 'llm', **kwargs) -> str:
        if scope is None:
            s = ''
            for scope in self:
                s += self._get_content_as_formatted_text(level, view_purpose, scope=scope, style=style)
        else:
            s = self.pretty_repr_for_scope_and_query(scope=scope, style=style,
                                                     **self.scopes_to_search_params[scope].to_dict())
        return s

    def _get_content_as_html(self, level: int, scope: Optional[str] = None, style: str = 'html', **kwargs) -> str:
        if scope is None:
            s = 'We searched for papers in the following scopes:\n'
            for scope in self:
                s += f'<h{level + 1}>{scope.title()}-related papers</h{level + 1}>\n'
                s += self._get_content_as_html(level, scope=scope, style=style)
        else:
            s = self.pretty_repr_for_scope_and_query(scope=scope, style=style,
                                                     **self.scopes_to_search_params[scope].to_dict())
        return s



from typing import List, Mapping, Any

import requests
from unidecode import unidecode

from .base_server import ParameterizedQueryServerCaller
from .custom_types import Citation
from .types import ServerErrorException

CROSSREF_URL = "https://api.crossref.org/works"

HEADERS = {
    "User-Agent": "data_to_paper (mailto:fallpalapp@gmail.com)"
}

BIBTEX_TEMPLATE = '@{type}{{{id},\n{fields}}}\n'

TYPE_MAPPING = {
    'journal-article': 'article',
    'proceedings-article': 'inproceedings',
    'book': 'book',
    'edited-book': 'inbook',
    'book-chapter': 'inbook',
}


def get_type_from_crossref(crossref: Mapping[str, Any]) -> str:
    return TYPE_MAPPING.get(crossref['type'], 'misc')


PAPER_MAPPING = {
                'authors': 'author',
                'title': 'title',
                'journal': 'journal',
                'year': 'year',
                'volume': 'volume',
                'issue': 'number',
                'pages': 'pages',
                'DOI': 'doi',
            }

BOOK_MAPPING = {
                'authors': 'author',
                'editors': 'editor',
                'title': 'title',
                'publisher': 'publisher',
                'year': 'year',
                'volume': 'volume',
                'issue': 'number',
                'pages': 'pages',
                'DOI': 'doi',
                'ISBN': 'isbn',
            }


STRS_TO_REMOVE_FROM_BIBTEX_ID = ["-", "_", "–", "’", "'", "/", " ", "(", ")", "[", "]", "{", "}", ":", ";",
                                 ",", ".", "?", "!", "“", "”", '"', "–"]


class CrossrefCitation(Citation):
    """
    A single crossref citation. This the raw dict after transforming the json response.
    This dict is hashable.
    """

    @property
    def bibtex_type(self) -> str:
        return get_type_from_crossref(self)

    @property
    def bibtex(self) -> str:
        # create a mapping for article and inproceedings
        bibtex_type = get_type_from_crossref(self)
        is_paper = bibtex_type in ['article', 'inproceedings']
        field_mapping = PAPER_MAPPING if is_paper else BOOK_MAPPING

        fields = []
        for key, value in self.items():
            if key in field_mapping:
                if value and key not in ['doi', 'isbn']:
                    # remove special characters of the value
                    if isinstance(value, list):
                        value = [unidecode(v).replace(r' &', r' \&').replace(r'_', r'\_') for v in value]
                    elif isinstance(value, str):
                        value = unidecode(value).replace(r' &', r' \&').replace(r'_', r'\_')
                bibtex_key = field_mapping[key]
                fields.append(f"{bibtex_key} = {{{value}}}")
        return BIBTEX_TEMPLATE.format(type=self.bibtex_type, id=self.bibtex_id, fields=',\n'.join(fields))

    @property
    def bibtex_id(self) -> str:
        """
        Get the bibtex id for this citation.
        """
        bibtex_id = unidecode(self['first_author_family']) + (str(self.get("year")) if self.get("year") else "")
        bibtex_id += self['title'].split(" ")[0] if self.get("title") else ""

        # remove special characters from the id like -, _, etc.
        for char in STRS_TO_REMOVE_FROM_BIBTEX_ID:
            bibtex_id = bibtex_id.replace(char, "")
        return bibtex_id


class CrossrefServerCaller(ParameterizedQueryServerCaller):
    """
    Search for citations in Crossref.
    """
    name = "Crossref"
    file_extension = "_crossref.bin"

    @staticmethod
    def crossref_item_to_citation(item):
        # create authors as a string in the format
        # "first_name1 last_name1 and first_name2 last_name2 and first_name3 last_name3"
        authors_string = ""
        for author in item["author"]:
            if author != item["author"][-1]:
                authors_string += f"{author.get('given', '')} {author.get('family', '')} and "
            else:
                authors_string += f"{author.get('given', '')} {author.get('family', '')}"

        # if editors are present, add them to the same way as authors
        editor_string = ""
        if item.get("editor", None) is not None:
            for editor in item["editor"]:
                if editor != item["editor"][-1]:
                    editor_string += f"{editor.get('given', '')} {editor.get('family', '')} and "
                else:
                    editor_string += f"{editor.get('given', '')} {editor.get('family', '')}"
        citation = {
            "title": item["title"][0],
            "first_author_family": item["author"][0]["family"].split(" ")[0],
            "authors": authors_string,
            "journal": item.get("container-title", [None])[0],
            "doi": item.get("DOI", ''),
            "type": item.get("type", ''),
            "year": item["published"]["date-parts"][0][0] if "published" in item else
            item["published-print"]["date-parts"][0][0] if "published-print" in item else '',
            "publisher": item.get("publisher", ''),
            "volume": item.get("volume", ''),
            "issue": item.get("issue", ''),
            "page": item.get("page", ''),
            "editors": editor_string if item.get("editor", None) is not None else '',
            "isbn": item.get("ISBN", '')
        }
        for key, value in citation.items():
            if isinstance(value, str) and "&NA;" in value:
                raise ValueError(f"Value {value} for key {key} is not valid")
        return citation

    @classmethod
    def _get_server_response(cls, query, rows=4) -> List[dict]:
        """
        Get the response from the crossref server as a list of CrossrefCitation objects.
        """
        params = {
            "query.bibliographic": query,
            "rows": rows,
            "filter": "type:journal-article,type:proceedings-article,type:book,type:edited-book,type:book-chapter",
            "select": "title,author,container-title,DOI,type,published,published-print,publisher,volume,issue,page,"
                      "editor,ISBN",
        }

        response = requests.get(CROSSREF_URL, headers=HEADERS, params=params)

        if response.status_code != 200:
            raise ServerErrorException(server=cls.name, response=response)

        data = response.json()
        items = data['message']['items']
        citations: List[dict] = []

        for item in items:
            try:
                citation = CrossrefServerCaller.crossref_item_to_citation(item)
            except (KeyError, ValueError):
                continue
            citations.append(citation)

        return citations

    @staticmethod
    def _post_process_response(response: List[dict], args, kwargs) -> List[CrossrefCitation]:
        """
        Post process the response from the server.
        """
        query = args[0] if args else kwargs.get("query", None)
        return [CrossrefCitation(citation, search_rank=rank, query=query) for rank, citation in enumerate(response)]


CROSSREF_SERVER_CALLER = CrossrefServerCaller()



from dataclasses import dataclass, field
from functools import partial
from typing import Dict, List, Collection, Optional, Any

from data_to_paper.env import PAUSE_AFTER_LITERATURE_SEARCH, JSON_MODE

from data_to_paper.utils import dedent_triple_quote_str, word_count
from data_to_paper.utils.nice_list import NiceDict, NiceList
from data_to_paper.utils.print_to_file import print_and_log_red
from data_to_paper.servers.semantic_scholar import SEMANTIC_SCHOLAR_SERVER_CALLER, \
    SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER

from data_to_paper.interactive import PanelNames

from .request_python_value import PythonDictWithDefinedKeysReviewBackgroundProductsConverser
from .literature_search import LiteratureSearch, QueryCitationCollectionProduct, \
    LiteratureSearchQueriesProduct


@dataclass
class BaseLiteratureSearchReviewGPT(PythonDictWithDefinedKeysReviewBackgroundProductsConverser):
    json_mode: bool = JSON_MODE
    number_of_papers_per_query: int = 100
    max_reviewing_rounds: int = 0
    literature_search: LiteratureSearch = field(default_factory=LiteratureSearch)
    domains_to_definitions_and_examples = {
        'dataset': {
            'definition': 'papers that use the same or similar datasets as in our study',
            'examples': ['The UK-NCD dataset', 'covid-19 vaccine efficacy dataset']
        },
        'questions': {
            'definition': 'papers that ask questions similar to our study',
            'examples': ['covid-19 vaccine efficacy over time', 'covid-19 vaccine waning']
        },
        'background': {
            'definition': 'papers that provide background on the overall subject of our study',
            'examples': ["SARS-CoV2 spread", "covid-19 global impact", "covid-19 vaccine"],
        },
        'methods': {
            'definition': 'papers that use the same or similar methods as in our study',
            'examples': ["covid-19 vaccine efficacy analysis", "kaplan-meier survival analysis"],
        },
        'results': {
            'definition': 'papers that report results similar to our study',
            'examples': ["covid-19 vaccine efficacy", "covid-19 vaccine efficacy over time", "covid-19 vaccine waning"],
        }
    }
    requested_keys: Collection[str] = ('dataset', 'questions',)

    excluded_citation_titles: List[str] = None
    # bibtex ids to remove from the results. Important for removing reproduced paper from citing itself.

    value_type: type = Dict[str, List[str]]
    goal_noun: str = 'literature search queries'
    goal_verb: str = 'write'

    mission_prompt: str = dedent_triple_quote_str("""
        Please write literature-search queries that we can use to search for papers related to our study.

        You would need to compose search queries to identify prior papers covering these {num_scopes} areas:
        {pretty_scopes_to_definitions}

        Return your answer as {your_response_should_be_formatted_as}, \t
        where the keys are the {num_scopes} areas noted above, \t
        and the values are lists of query string.

        Each individual query should be a string with up to 5-10 words. 

        For example, for a study reporting waning of the efficacy of the covid-19 BNT162b2 vaccine based on analysis \t
        of the "United Kingdom National Core Data (UK-NCD)", the queries could be:
        ```{python_or_json}
        {pretty_scopes_to_examples}
        ```

        Your response should be formatted as {your_response_should_be_formatted_as}  
        """)

    @property
    def chosen_domains_to_definitions_and_examples(self) -> Dict[str, Dict[str, str]]:
        return {key: self.domains_to_definitions_and_examples[key] for key in self.requested_keys}

    @property
    def pretty_scopes_to_definitions(self) -> str:
        return '\n'.join([f'"{scope}": {definition_and_examples["definition"]}'
                          for scope, definition_and_examples
                          in self.chosen_domains_to_definitions_and_examples.items()])

    @property
    def pretty_scopes_to_examples(self) -> str:
        nice = partial(NiceList, wrap_with='"', prefix='[', suffix=']', separator=', ')
        return ('{\n' +
                '\n'.join([f'    "{scope}": {nice(definition_and_examples["examples"])}'
                           for scope, definition_and_examples
                           in self.chosen_domains_to_definitions_and_examples.items()]) +
                '\n}')

    @property
    def num_scopes(self) -> int:
        return len(self.requested_keys)

    def get_title(self) -> Optional[str]:
        """
        Returns the title of the paper we are writing.
        None to skip building embedding vector.
        """
        return None

    def get_abstract(self) -> Optional[str]:
        """
        Returns the abstract of the paper we are writing.
        None to skip building embedding vector.
        """
        return None

    def _check_response_value(self, response_value: dict) -> NiceDict:
        super()._check_response_value(response_value)
        # The queries are 'valid' even if they have too many words:
        self._update_valid_result(response_value)
        too_long_queries = []
        for queries in response_value.values():
            for query in queries:
                if word_count(query) > 10:
                    too_long_queries.append(query)
        if too_long_queries:
            self._raise_self_response_error(
                title='# Too long queries',
                error_message=dedent_triple_quote_str("""
                Queries should be 5-10 word long.

                The following queries are too long:
                {}

                Please return your complete response again, with these queries shortened.
                """).format(NiceList(too_long_queries, wrap_with='"', prefix='', suffix='', separator='\n')))
        return NiceDict({k: NiceList(v, wrap_with='"', prefix='[\n' + ' ' * 8, suffix='\n' + ' ' * 4 + ']',
                                     separator=',\n' + ' ' * 8)
                         for k, v in response_value.items()})

    def _send_html_and_scroll_to_bottom(self, html: str):
        self._app_send_prompt(PanelNames.FEEDBACK, html, provided_as_html=True, scroll_to_bottom=True)

    def get_literature_search(self) -> LiteratureSearch:
        scopes_to_list_of_queries = self.run_and_get_valid_result()
        literature_search = self.literature_search
        with self._app_temporarily_set_panel_status(PanelNames.FEEDBACK, 'Querying citations...'):
            html = '<h1>Literature Search</h1>\n'
            if self.get_title() is not None and self.get_abstract() is not None:
                # Requesting embedding vector of our paper
                html += f'<h2>Getting Embedding Vector</h2>\n'
                html += f'<p>Requesting "{SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER.name}" ' \
                        f'for the embedding vector of our draft title and abstract....</p>\n'
                self._send_html_and_scroll_to_bottom(html)
                literature_search.embedding_target = \
                    SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER.get_server_response({
                        "paper_id": "",
                        "title": self.get_title(),
                        "abstract": self.get_abstract()})
                html += '<p>Embedding vector successfully retrieved.</p>\n' \
                        '<p>Now we can sort the papers by similarity to our study.</p>\n'
                self._send_html_and_scroll_to_bottom(html)

            html += f'<h2>Querying Citations</h2>\n'
            html += f'<p>Searching "{SEMANTIC_SCHOLAR_SERVER_CALLER.name}" ' \
                    f'for papers related to our study in the following areas:</p>\n'
            for scope, queries in scopes_to_list_of_queries.items():
                queries_to_citations = {}
                html += f'<h3>{scope.title()}-related queries:</h3>\n'
                for query in queries:
                    citations = SEMANTIC_SCHOLAR_SERVER_CALLER.get_server_response(query,
                                                                                   rows=self.number_of_papers_per_query)
                    num_citations = len(citations)
                    html += f'<p><b style="color: #1E90FF;">Query:</b> "{query}".\n'
                    html += f'<br><b style="color: #1E90FF;">Found:</b> {num_citations} citations.</p>\n'
                    self._send_html_and_scroll_to_bottom(html)
                    self.comment(f'\nQuerying Semantic Scholar. '
                                 f'Found {num_citations} / {self.number_of_papers_per_query} citations. '
                                 f'Query: "{query}".')
                    if self.excluded_citation_titles is not None:
                        excluded_citations = [citation for citation in citations
                                              if citation['title'] in self.excluded_citation_titles]
                        if excluded_citations:
                            print_and_log_red(
                                f'The following citations specified in the excluded citation list were excluded:\n')
                            for citation in excluded_citations:
                                print_and_log_red(f'{citation}\n\n')
                            citations = QueryCitationCollectionProduct(
                                query=query,
                                value=[citation for citation in citations if citation not in excluded_citations])

                    queries_to_citations[query] = citations

                literature_search[scope] = queries_to_citations

        self._app_request_panel_continue(PanelNames.FEEDBACK, sleep_for=PAUSE_AFTER_LITERATURE_SEARCH)
        return literature_search

    def _update_valid_result(self, valid_result: Any):
        super()._update_valid_result(LiteratureSearchQueriesProduct(value=valid_result))



from .highlighted_text import format_text_with_code_blocks
from .text_counting import line_count, word_count
from .text_extractors import extract_text_between_tags, extract_text_between_brackets
from .text_formatting import dedent_triple_quote_str



from abc import abstractmethod, ABCMeta
from dataclasses import dataclass, is_dataclass, fields
from typing import Optional, Union

from data_to_paper.conversation.stage import Stage


class data_to_paperException(Exception, metaclass=ABCMeta):
    """
    Base class for all exceptions in this package.
    """
    @abstractmethod
    def __str__(self):
        pass

    def __reduce__(self):
        if is_dataclass(self):
            # Collect the field values in the order they're defined in the dataclass
            field_values = [getattr(self, f.name) for f in fields(self)]
            return (self.__class__, tuple(field_values))
        else:
            # Fallback for non-dataclass exceptions
            return super().__reduce__()


class TerminateException(data_to_paperException, metaclass=ABCMeta):
    """
    Base class for all exceptions that terminate data-to-paper run.
    """
    pass


@dataclass
class MissingInstallationError(TerminateException):
    """
    An exception to indicate a missing installation.
    """
    package_name: str
    instructions: Optional[str] = None

    def __str__(self):
        s = f"{self.package_name} is not installed. Please install it."
        if self.instructions:
            s += f"\n\n{self.instructions}"
        return s


@dataclass
class ResetStepException(Exception):
    """
    An exception to reset the step.
    """
    stage: Optional[Union[Stage, bool]]  # reset to this stage, None to exit the app



import signal
import threading
import os
from dataclasses import dataclass
from typing import Type

from data_to_paper.run_gpt_code.base_run_contexts import RegisteredRunContext


def timeout_context(seconds, exception=TimeoutError):
    """
    Context manager to terminate a function if runtime is too long.
    """

    # return different context manager depending on the operating system
    if os.name == 'nt':
        return TimeoutWindowsContext(seconds=seconds, exception=exception) # TODO: currently not working
    else:
        return TimeoutUnixContext(seconds=seconds, exception=exception)


@dataclass
class BaseTimeoutContext(RegisteredRunContext):
    seconds: int = 10
    exception: Type[Exception] = TimeoutError


@dataclass
class TimeoutUnixContext(BaseTimeoutContext):

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.signal_handler)
        signal.alarm(self.seconds)
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.alarm(0)
        return super().__exit__(exc_type, exc_val, exc_tb)

    def signal_handler(self, signum, frame):
        raise self.exception(self.seconds)


@dataclass
class TimeoutWindowsContext(BaseTimeoutContext):

    def __enter__(self):
        self.stop_event = threading.Event()
        self.worker_thread = threading.Thread(target=self.target)
        self.worker_thread.start()
        self.worker_thread.join(timeout=self.seconds)
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.stop_event.is_set():
            raise self.exception(self.seconds)
        return super().__exit__(exc_type, exc_val, exc_tb)

    def target(self):
        try:
            yield
        finally:
            self.stop_event.set()



import re
from typing import List


def find_citation_ids(latex_string) -> List[str]:
    """
    Find all the citation ids in the latex string.

    For example, if the latex string is
    "one citation \\cite{citation_id1} two citations \\cite{citation_id2, citation_id3}", then the function will return
    ['citation_id1', 'citation_id2', 'citation_id3'].
    """
    pattern = r"\\cite{([^}]+)}"
    matches = re.findall(pattern, latex_string)
    citation_ids = [citation_id.strip() for match in matches for citation_id in match.split(',')]
    return citation_ids


def choose_first_citation(sentence_citations):
    """
    Choose the first citation for the sentence, if any.
    """
    chosen_citations_ids = [sentence_citations[0]['bibtex'].split('{')[1].split(',\n')[0]]
    chosen_citations_indices = [0]
    return chosen_citations_ids, chosen_citations_indices


def remove_tags(text):
    text = re.sub(r'<[^>]+>', '', text)  # Remove HTML tags
    text = re.sub(r'&lt;|&gt;', '', text)  # Remove &lt; and &gt;
    text = re.sub(r'^Abstract', 'Abstract ', text)  # Add space after "Abstract" at the beginning
    text = re.sub(r'^p|/p$', '', text)  # Remove "p" and "/p" at the beginning and end
    return text.strip()  # Remove leading and trailing whitespace
